/* pcap-related interfacey functions here... */
#include <string.h>
#include <stdio.h>
#include <sys/time.h>
#include <unistd.h>
#include <dnet.h>

#include "lsrscan.h"

int getLinkSize(pcap_t *pcapPtr) {
  int linkType;
  
  linkType = pcap_datalink(pcapPtr);
  
  if(linkType == DLT_NULL) return(4);
  else if (linkType == DLT_EN10MB) return(14);
  else if (linkType == DLT_IEEE802) return(22);
  else if (linkType == DLT_RAW) return (0);
  else if (linkType == DLT_FDDI) return (21);
  else return(-1); /* Unsupported as of yet...*/
} 

pcap_t * initListen(struct packetInfo *settings, struct addr *listenAddr) {
  char errbuf[PCAP_ERRBUF_SIZE];
  char listenAddrString[FQDNMAX];
  bpf_u_int32 myAddr, myNetmask;
  char *defaultInterface;
  pcap_t *pcapPtr;
  intf_t *localInterface;
  struct intf_entry *intfEntry;
  int ret;
  char filter[FILTERSIZE];
  uint8_t buf[TMPBUFSIZE];

  struct bpf_program filterCode;

  /* 
   * Note, the following works fine on machines with one interface 
   * I'm going to have to go back and combine route_get() with 
   * pcap_findalldevs() to make it bullet-resistant 
   */

  if((defaultInterface = pcap_lookupdev(errbuf)) == NULL) {
    fprintf(stderr, "pcap_lookupdev failed: %s\n", errbuf);
    return(NULL);
  }

  if((localInterface=intf_open()) == NULL) {
    fprintf(stderr, "intf_open failed\n");
    return(NULL);
  }
  
  if(settings->silentIP == 0) {
    listenAddr->addr_type = ADDR_TYPE_IP;
    memset(buf, 0, TMPBUFSIZE);
    intfEntry = (struct intf_entry *)buf;    
    strncpy(intfEntry->intf_name, defaultInterface, 59);
    intfEntry->intf_len = TMPBUFSIZE;
    
    if(intf_get(localInterface, intfEntry) < 0) {
	fprintf(stderr, "intf_get failed\n");
	intf_close(localInterface);
	return(NULL);
    }
    intf_close(localInterface);
  }  

  memcpy(&(listenAddr->addr_ip), &(intfEntry->intf_addr.addr_ip), IP_ADDR_LEN);

  listenAddr->addr_bits = 32;

  if((addr_ntop(listenAddr, listenAddrString, FQDNMAX)) < 0 ) {
    fprintf(stderr, "addr_ntop: listenAddr conversion failed\n", listenAddr->addr_ip);
    return(NULL);
  }

  if((pcap_lookupnet(defaultInterface, &myAddr, &myNetmask, errbuf)) < 0) {
    fprintf(stderr, "pcap_lookupnet failed: %s\n", errbuf);
    return(NULL);
  }  

  if((pcapPtr = pcap_open_live(defaultInterface, PACKETSIZE, 1, PACKETWAIT,
			       errbuf)) == NULL) {
    fprintf(stderr, "pcap_open_live failed: %s\n",errbuf);
    return(NULL);
  }

  if((pcap_setnonblock(pcapPtr, 1, errbuf)) < 0) {
    fprintf(stderr, "pcap_setnonblock failed: %s\n",errbuf);
    pcap_close(pcapPtr);
    return(NULL);
  } 

  snprintf(filter, sizeof(filter)-1, "tcp and port %d and host %s",
	   settings->destPort, listenAddrString); 
  
  if ((pcap_compile(pcapPtr, &filterCode, filter, 1, myNetmask)) < 0) {
    fprintf(stderr, "pcap_compile failed: %s\n", pcap_geterr(pcapPtr));
    pcap_close(pcapPtr);
    return(NULL);
  }
  
  ret = pcap_setfilter(pcapPtr, &filterCode);
  pcap_freecode(&filterCode);
  if(ret < 0) {
    fprintf(stderr, "pcap_setfilter failed: %s\n", pcap_geterr(pcapPtr));
    pcap_close(pcapPtr);
    return(NULL);
  }
  return(pcapPtr);
}

int checkPacket(uint8_t *packet, struct packetInfo *settings) {
  struct ip_hdr *ipHeader;
  struct tcp_hdr *tcpHeader;
  int i;
  
  ipHeader = (struct ip_hdr *)packet;
  if(ipHeader->ip_p == IPPROTO_TCP)
    tcpHeader = (struct tcp_hdr *)(packet+(4*ipHeader->ip_hl));
  else 
    return(0);

  if(settings->type == TYPE_THROUGH) {
    if((htonl(tcpHeader->th_seq) == htonl(settings->sequenceNum)) &&
       (ipHeader->ip_src == settings->ipHops[0]) &&
       (ipHeader->ip_dst == settings->ipHops[settings->myHop]) &&
       (tcpHeader->th_sport == htons(settings->sourcePort)) &&
       (tcpHeader->th_dport == htons(settings->destPort)))
      return(1);
  }
  if(settings->type == TYPE_TO) {
    if((htonl(tcpHeader->th_ack) == (htonl(settings->sequenceNum) + 1)) &&
       (ipHeader->ip_src == settings->ipHops[settings->hopCount-1]) &&
       (ipHeader->ip_dst == settings->ipHops[settings->myHop]) &&
       (tcpHeader->th_sport == htons(settings->destPort)) &&
       (tcpHeader->th_dport == htons(settings->sourcePort)))
      return(1);
  }
  return(0);
}

int getResponse(pcap_t *pcapPtr, struct packetInfo *settings) {
  struct timeval start,current;
  uint8_t *packet;
  int retVal;
  int linkHeadSize;
  struct pcap_pkthdr packetHead;

  if((linkHeadSize = getLinkSize(pcapPtr)) < 0) {
    fprintf(stderr, "Unsupported link type\n");
    return(-1);
  }

  gettimeofday(&start, NULL);
  gettimeofday(&current, NULL);

  while((((current.tv_sec - start.tv_sec) * 1000) + 
	 ((current.tv_usec - start.tv_usec) / 1000)) < TIMEOUT) {
    packet = NULL;
    while((packet = (uint8_t *)pcap_next(pcapPtr, &packetHead)) != NULL) {
      packet += linkHeadSize;
      retVal = checkPacket(packet, settings);
      if(retVal == 1) return(1);
    }
    usleep(PACKETWAIT*1000);
    gettimeofday(&current, NULL);
  }
  return(0); 
}
